/*
 * $Id: active_probing.c,v 1.5 2001/07/24 23:29:16 laik Exp $
 *
 * Copyright (C) 1998 Kevin Lai
 *
 * This module contains code for doing active probing.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>

#include "kl_array.h"

#include "active_probing.h"
#include "transport.h"

struct ActiveProbing {
    ActiveProbingDef def;              /* Definitions. */
    KLEventDef timer_def;
    KLEventDefRegistration *timer_reg; /* For the timer. */

    TransportSession *transport_session;/* Used to send/receive packets. */
    TransportBuffer **transport_bufs;

    unsigned int packets_sent;         /* # packets sent */
    unsigned int packets_received;     /* # packets received */
    unsigned int packet_timeouts;      /* timeouts */
    unsigned int last_burst_received;
    FilterResult result;
};

KLExceptionClass active_probing_below_max_error_ec = {
    "Below maximum active probing error."
};

KLExceptionClass ap_exceeded_sent_packets_maximum[] = {
    {"Exceeded active probing sent packets maximum."}
};

/*
 * Finish active probing.
 */
static void
ap_finish(ActiveProbing *ap)
{
    printf("BW: %f Error: %f Samples: %d\n", 
	   ap->result.bw, ap->result.error, ap->result.num_samples);
    printf("Active probing statistics:\n");
    printf("  packets sent: %d \n", ap->packets_sent);
    printf("  packets received: %d\n", ap->packets_received);
    printf("  timeouts: %d\n", ap->packet_timeouts);
}

/*
 * Determine if this packet is one of ours.
 */
static unsigned int
ap_check_packet(
    const DPCapFlowPacket *packet, int ack, struct timespec *delay)
{
    const DPCapFlowPacket *ack_packet;
    const DPCapFlow *dpcap_flow;
    struct timespec *timings, *ack_timings;
    DPCapFlowDirection direction;
    const DPCapFlowSet *flow_set;
    int num_timers;

    dpcap_flow = DPCapFlowPacketFlow(packet);
    timings = DPCapFlowPacketTimings(packet);
    direction = DPCapFlowGetDirection(dpcap_flow);
    flow_set = DPCapFlowGetSet(dpcap_flow);
    num_timers = DPCapFlowSetNumTimers(flow_set);
    /* Compute round trip or one-way delay. */
    KLTimingSet(delay, 0, 0);
    if (ack) {
	ack_packet = DPCapFlowPacketCumAck(packet);
	if (ack_packet == NULL) {
	    ack_packet = DPCapFlowPacketExactAck(packet);
	}
	ack_timings = DPCapFlowPacketTimings(ack_packet);

	/*
	 * Get delay when we get an ack for a packet that
	 * we sent.
	 */
	if ((direction == DPCapFlowDirectionForward) && 
	    ((KLTimingGetSeconds(&timings[0]) != 0) &&
	     (KLTimingGetNS(&timings[0]) != 0)) && 
	    ((KLTimingGetSeconds(&ack_timings[0]) != 0) &&
	     (KLTimingGetNS(&ack_timings[0]) != 0))) {
	    KLTimingSubtract(delay, &ack_timings[0], &timings[0]);
	    KLTimingNormalize(delay);
	    return 2;
	}
    }
    else {
	/*
	 * Get delay when we have the source and destination timings
	 * of a packet.
	 */
	if ((direction == DPCapFlowDirectionForward) && 
	    ((KLTimingGetSeconds(&timings[0]) != 0) &&
	     (KLTimingGetNS(&timings[0]) != 0)) && 
	    ((KLTimingGetSeconds(&timings[num_timers - 1]) != 0) &&
	     (KLTimingGetNS(&timings[num_timers - 1]) != 0))) {
	    KLTimingSubtract(delay, &timings[num_timers - 1], &timings[0]);
	    KLTimingNormalize(delay);
	    return 1;
	}
    }
    return 0;
}

/*
 * 
 */
static KLException *
ap_send_packets(ActiveProbing *ap)
{
    KLException *ex = NULL;
    unsigned int b, new_packet_size = 0;

    /* Check if we can send any more packets. */
    if (ap->packets_sent >= ap->def.max_packets_sent) {
	ap_finish(ap);
	return KLExceptionThrow(ap_exceeded_sent_packets_maximum);
    }

    /* Send packets. */
    for (b = 0; b < ap->def.burst_size; b++) {
	transport_compose_packet(
	    ap->transport_session, &ap->def.flow.dest.sock_addr,
	    IPPROTO_TCP, 255, ap->def.packet_size, b + ap->packets_sent, 
	    ap->transport_bufs[b]);
    }
    for (b = 0; b < ap->def.burst_size; b++) {
	new_packet_size = transport_send_packet(
	    ap->transport_session, ap->transport_bufs[b]);	
    }
    ap->packets_sent += ap->def.burst_size;
    ap->last_burst_received = 0;

    /* Check for path MTU change. */
    if (new_packet_size != ap->def.packet_size) {
	printf("There was a path MTU change\n.");
	ap->def.packet_size = new_packet_size;
    }

    /* Reset timeout. */
    KLEventDefUnregister(ap->def.manager, ap->timer_reg);
    ap->timer_reg = KLEventDefRegister(ap->def.manager, &ap->timer_def);

    return ex;
}

/*
 * This function handles the expiration of the sending timer event.
 * The timer prevents us from waiting too long before sending the
 * next packet.
 */
static KLException *
ap_timer_event_handler(const KLEvent *event)
{
    ActiveProbing *ap = event->def->client_data;

    printf("Burst timed out. Trying again.\n");
    ap->packet_timeouts++;
    return ap_send_packets(ap);
}

/*
 * Allocate and initialize new state.
 */
ActiveProbing *
active_probing_new(ActiveProbingDef *def)
{
    ActiveProbing *ap;
    TransportSessionDef ts_def;
    unsigned int min_packet_size, max_packet_size;
    unsigned int b;

    ap = malloc(sizeof(ActiveProbing));
    memcpy(&ap->def, def, sizeof(ActiveProbingDef));

    ap->packets_sent = 0;
    ap->packets_received = 0;
    ap->packet_timeouts = 0;
    ap->last_burst_received = 0;
    ap->result.error = 100.0;

    /* Set up packet sending state. */
    memcpy(
	&ts_def.source, &ap->def.flow.source, sizeof(TransportAddress));
    memcpy(
	&ts_def.dest, &ap->def.flow.dest, sizeof(TransportAddress));
    KLExceptionErrorCatch(transport_new_session(
	&ts_def, &ap->transport_session), "");
    ap->transport_bufs = malloc(
	sizeof(TransportBuffer *) * ap->def.burst_size);
    for (b = 0; b < ap->def.burst_size; b++) {
	ap->transport_bufs[b] = transport_new_buffer(
	    ap->transport_session);	
    }

    transport_session_get_packet_sizes(
	ap->transport_session, IPPROTO_TCP, &min_packet_size,
	&max_packet_size);
    ap->def.packet_size = min(max_packet_size, ap->def.packet_size);
    ap->def.packet_size = max(min_packet_size, ap->def.packet_size);

    /* Register sending function on a timer. */
    ap->timer_def.type = KLEventTypeTimer;
    ap->timer_def.client_data_delete_fun = NULL;
    ap->timer_def.client_data = ap;
    ap->timer_def.handler_fun = ap_timer_event_handler;
    KLTimingCopy(
	&ap->timer_def.type_data.timer_interval, &ap->def.timeout);
    ap->timer_reg = KLEventDefRegister(ap->def.manager, &ap->timer_def);

    return ap;
}

/*
 * Free resources associated with a link bandwidth calculation state.
 */
void
active_probing_free(ActiveProbing *ap)
{
    unsigned int b;

    KLEventDefUnregister(ap->def.manager, ap->timer_reg);
    for (b = 0; b < ap->def.burst_size; b++) {
	transport_free_buffer(ap->transport_bufs[b]);
    }
    free(ap->transport_bufs);
    transport_delete_session(ap->transport_session);
    free(ap);
}

/*
 * 
 */
KLException *
active_probing_start(ActiveProbing *ap)
{
    return ap_send_packets(ap);
}

/*
 * Called when new information about a captured packet arrives.
 */
KLException *
active_probing_packet_new(
    FlowCalcFlow *flow, const KLListNode *cur_node, int timer_index,
    int ack, void *flow_set_client_data)
{
    ActiveProbing *ap = flow_set_client_data;
    struct timespec delay;
    const DPCapFlowPacket *packet;
    unsigned int timing_type = 1, have_result = 0;
    KLException *ex = NULL;
    unsigned short packet_id;
    const u_int8_t *payload;
    const KLNetworkProtocolPacket *np_p;
    const char *metric_name;

    packet = KLListGetItem(cur_node);
    timing_type = ap_check_packet(packet, ack, &delay);

    /* Only care about full timings. */
    if (timing_type != 2) {
	return 0;
    }

    /* Check if this sample is one of ours. */
    np_p = DPCapFlowPacketGetPacket(packet);
    payload = transport_check_packet(ap->transport_session, np_p);
    if (payload == NULL) {
	return 0;
    }
    ap->packets_received++;
    ap->last_burst_received++;
    packet_id = transport_get_packet_id(
	ap->transport_session, &ap->def.flow.dest.sock_addr,
	np_p);
    // printf("packet_id: %d\n", packet_id);

    /* Get latest estimate. */
    have_result = flow_calc_get_metrics(flow, 0, 0, &metric_name, &ap->result);
    // printf("have_result: %d\n", have_result);

#if 0
    if (have_result) {
	printf("BW: %f Error: %f # Samples: %d\n", 
	       ap->result.bw, ap->result.error, ap->result.num_samples);
    }
#endif

    /* Finish if we have a sufficiently good estimate. */
    if ((ap->result.num_samples >= ap->def.min_samples) &&
	(ap->result.error < ap->def.max_error)) {
	ap_finish(ap);
	ex = KLExceptionThrow(&active_probing_below_max_error_ec);
    }
    /* Finish if we have sent enough packets. */
    else if ((ap->packets_sent > ap->def.max_packets_sent) ||
	     (ap->packets_received >= ap->def.max_packets_sent)) {
	ap_finish(ap);
 	ex = KLExceptionThrow(ap_exceeded_sent_packets_maximum);
    }
    /* if we got all of this burst, send another one. */
    else if (ap->last_burst_received == ap->def.burst_size) {
	ap_send_packets(ap);
    }
    return ex;
}

