/*
 * $Id: calc.h,v 1.9 2001/03/30 22:54:35 laik Exp $
 *
 * This file contains information about the simulation routines.
 *
 * Copyright (C) 1998 Kevin Lai
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __CALC_H__
#define __CALC_H__

/* Local includes */

#include "kl_buffered_output.h"
#include "kl_list.h"
#include "kl_statistics_kernel.h"

struct PerfCalc;
typedef struct PerfCalc PerfCalc;

#define CALC_BUNCH_SIZE_MAX 8

/*
 * These are all the performance metrics that we are interested in.
 */
typedef enum CalcMetric {
    CalcMetricSBBW = 1,           /* Sender-Based Bandwidth */
    CalcMetricRBBW = 2,           /* Receiver-Based Bandwidth */
    CalcMetricROBW = 3,           /* Receiver-Only Bandwidth */
    CalcMetricPBW = 4,            /* Potential Bandwidth */
    CalcMetricABW = 5,            /* Available Bandwidth */
    CalcMetricRTT = 6,            /* Round-trip time */
    CalcMetricArrivalTime = 7,    /* Arrival time. */
    CalcMetricDelay = 8,          /* One way delay */
    CalcMetricLast = 9            /* The last metric */
} CalcMetric;

/*
 * This structure encapsulates the specifications of how the bandwidth
 * calculation should be performed.
 */
typedef struct CalcDef {
    int min_bunch_size;           /* Size of bunch to use. */
    int max_bunch_size;           /* Size of bunch to use. */
    long restrict_packet_pair;/* Flags for interested metrics. */
    int window_size;          /* How many packets to use. */
    double min_delta_rtt;     /* The minimum deviation in delta rtt. */
    double max_delta_rtt;     /* The maximum deviation in delta rtt. */
    double no_bw_ratio;       /* The ratio below which we cannot find a bw. */
    double clock_resolution;  /* The clock resolution. */
    double kernel_bound_ratio;
    double kernel_resolution;
    double kernel_weight;
    double pbw_weight;
    double mbw_weight;
    int verbose;
    KLStatisticsKernelFunction kernel_function;
    KLBufferedOutput *sample_log;
} CalcDef;

typedef struct CalcMetricSpec {
    const char *name;
    long flag;
} CalcMetricSpec;

extern const CalcMetricSpec CalcMetricSpecs[];

/*
 * This structure encapsulates summary information about
 * what happened during the calculation.
 */
typedef struct PerfCalcSummaryMetrics {
    int not_same_size;    /* Bunches aborted that weren't the same size. */
    int compressed;       /* Bunches aborted because of compression. */
    int no_rtt;           /* Bunches aborted because of no rtt. */
    int num_bunches;
} PerfCalcSummaryMetrics;

/*
 * This structure encapsulates all information about a bandwidth sample.
 */
typedef struct PerfCalcSampleSet {
    long valid_metrics;
    double d[CalcMetricLast];    /* Meaning defined as above. */
} PerfCalcSampleSet;

int
perf_calc_new_packet(
    PerfCalc *perf_calc, const KLListNode *first_node,
    const KLListNode *cur_node, int send_index, int arrival_index,
    int ack_index);

PerfCalc *
perf_calc_new(CalcDef *calc_spec);

void
perf_calc_delete(PerfCalc *perf_calc);

void
perf_calc_verify(PerfCalc *perf_calc);

PerfCalcSampleSet *
perf_calc_get_sample_set(const PerfCalc *perf_calc, int *num_sets);

const PerfCalcSummaryMetrics *
perf_calc_get_summary_metrics(const PerfCalc *perf_calc);

#endif


