/*
 * $Id: client.c,v 1.27 2001/07/20 22:55:23 laik Exp $
 *
 * Copyright (C) 1998 Kevin Lai
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
#include <assert.h>
#include <string.h>
#include <unistd.h>

#include "kl_array.h"
#include "kl_string.h"
#include "kl_buffered_output.h"
#include "kl_socket.h"

#include "dpcap_client.h"

#include "active_probing.h"
#include "flow_calc.h"
#include "link_bandwidth_calc.h"

#include "client.h"

struct Client {
    FlowCalcSet *flow_calc_set;             /* Packet pair flows */
    KLBufferedOutput *bo;                   /* Update file */
    KLBufferedOutput *result_log;           /* Result log */
    KLBufferedOutput *sample_log;           /* Sample log */
    LinkBandwidthCalc *link_bandwidth_calc; /* Link bandwidth calculation */
    ActiveProbing *ap;                      /* Active probing */
    DPCapClient *dpcap_client;              /* Distributed packet capture */
};

const char *timers[2] = {"84", "83"};

static void
fill_calc_def(KLParseOptionValue *args, CalcDef *calc_def)
{
    calc_def->max_bunch_size = (unsigned int) args[ARG_CALC_MAX_BUNCH_SIZE].i;
    calc_def->min_bunch_size = (unsigned int) args[ARG_CALC_MIN_BUNCH_SIZE].i;
    calc_def->restrict_packet_pair = args[ARG_CALC_RESTRICT_PACKET_PAIR].i;
    calc_def->window_size = args[ARG_CALC_WINDOW_SIZE].i;
    calc_def->min_delta_rtt = args[ARG_CALC_MIN_DELTA_RTT].d;
    calc_def->max_delta_rtt = args[ARG_CALC_MAX_DELTA_RTT].d;
    calc_def->no_bw_ratio = args[ARG_CALC_NO_BW_RATIO].d;
    calc_def->clock_resolution = 0.00001;
    calc_def->kernel_bound_ratio = args[ARG_CALC_KERNEL_BOUND_RATIO].d;
    calc_def->kernel_resolution = args[ARG_CALC_KERNEL_RESOLUTION].d;
    calc_def->kernel_function = (KLStatisticsKernelFunction) 
	args[ARG_CALC_KERNEL_FUNCTION].a;
    calc_def->kernel_weight = args[ARG_CALC_KERNEL_WEIGHT].d;
    calc_def->pbw_weight = args[ARG_CALC_PBW_WEIGHT].d;
    calc_def->mbw_weight = args[ARG_CALC_MBW_WEIGHT].d;
    calc_def->verbose = args[ARG_VERBOSE].i;
}

/*
 * Initialize the update file. 
 */
static KLException *
c_output_log_init(
    KLEventManager *manager, KLParseOptionValue *args, char *log_name,
    KLBufferedOutput **bo_p)
{
    KLBufferedOutput *bo = NULL;
    KLBufferedOutputDef bo_def;
    KLException *ex;

    bo_def.manager = manager;
    bo_def.flush_flags = KLBufferedOutputFlushFlagSize;
    bo_def.flush_thresh = 0;
    bo_def.update = NULL;
    bo_def.new_client = NULL;
    bo_def.delete_client = NULL;
    bo_def.input = NULL;
    bo_def.verbosity = args[ARG_VERBOSE].i;

    /* If a file name specified, write to file. */
    if (log_name[0] != '-') {
	bo_def.type = KLBufferedOutputTypeFile;
	bo_def.type_data.file.file_name = log_name;
    }
    /* Write to stdout as if it were a file. */
    else {
	bo_def.type = KLBufferedOutputTypeFile;
	bo_def.type_data.file.fd = fileno(stdout);
	bo_def.type_data.file.file_name = "";
    }
    ex = KLBufferedOutputOpen(&bo_def, &bo);
    *bo_p = bo;
    return ex;
}

static void
c_active_probe_flow_init(KLParseOptionValue *args, TransportFlow *flow)
{
    KLParseOptionServers *servers = NULL;
    int s;
    char *src_name = NULL, *src_service = NULL, *dst_name, *dst_service;

    /* Get the dest/src address/service */
    servers = KLParseOptionServersParse(
	args[ARG_ACTIVE_PROBING_ADDRESSES].s, "10100");
    s = 0;
    if (KLParseOptionNumServers(servers) >= 2) {
	KLParseOptionGetServer(servers, s, &src_name, &src_service);
	s++;
    }
    KLParseOptionGetServer(servers, s, &dst_name, &dst_service);
    transport_flow_compose(
	flow, src_name, src_service, dst_name, dst_service);

    KLParseOptionServersFree(servers);
}

/*
 * Initialize the link bandwidth client.
 */
static KLException *
c_lbc_init(
    Client *client, KLEventManager *manager, KLParseOptionValue *args,
    DPCapClientDef *client_def, KLBufferedOutputDef *bo_def)
{
    LinkBandwidthCalc *lbc;
    LinkBandwidthCalcDef lbc_def;
    char address_string[INET_ADDRSTRLEN];
    KLException *ex = NULL;
    int len;
    KLArray *a;

    lbc_def.manager = manager;

    c_active_probe_flow_init(args, &lbc_def.flow);

    /* Set local capture. */
    a = KLArrayNew();
    KLArrayStrCat(a, "localhost");
    args[ARG_DPCAP_SERVERS].s = KLArrayGetBuffer(a, &len);
    KLArrayFree(a, 0);

    inet_ntop(
	AF_INET, 
	&((struct sockaddr_in *) &lbc_def.flow.source.sock_addr)->sin_addr,
	address_string, sizeof(address_string));
    // printf("Using source address %s\n", address_string);

    lbc_def.flow.ip_protocol = (int)
	args[ARG_ACTIVE_PROBING_PROTOCOL].a;
    lbc_def.sigma_min_packets = 
	args[ARG_ALL_LINK_BANDWIDTH_SIGMA_MIN_PACKETS].i;
    lbc_def.sigma_max_packets = 
	args[ARG_ALL_LINK_BANDWIDTH_SIGMA_MAX_PACKETS].i;
    lbc_def.sigma_min_confidence = 
	args[ARG_ALL_LINK_BANDWIDTH_SIGMA_MIN_CONFIDENCE].d;
    lbc_def.tailgate_min_packets = 
	args[ARG_ALL_LINK_BANDWIDTH_TAILGATE_MIN_PACKETS].i;
    lbc_def.tailgate_max_packets = 
	args[ARG_ALL_LINK_BANDWIDTH_TAILGATE_MAX_PACKETS].i;
    lbc_def.tailgate_max_error = 
	args[ARG_ALL_LINK_BANDWIDTH_TAILGATE_MAX_ERROR].d;
    KLTimingSetDouble(
	&lbc_def.timeout, args[ARG_ALL_LINK_BANDWIDTH_TIMEOUT].d);
    KLTimingSetDouble(
	&lbc_def.max_sending_interval, 
	args[ARG_ALL_LINK_BANDWIDTH_MAX_SENDING_INTERVAL].d);

    if (args[ARG_LOG_FILE_NAME].s[0] != '\0') {
	ex = c_output_log_init(
	    manager, args, args[ARG_LOG_FILE_NAME].s, &lbc_def.output_log);
	if (ex != NULL) {
	    return ex;
	}
    }

    lbc = link_bandwidth_calc_new(&lbc_def);

    bo_def->update = link_bandwidth_calc_update_output;
    bo_def->client_data = lbc;

    client_def->flow_set_def.packet_new_fun = link_bandwidth_calc_packet_new;
    client_def->flow_set_def.flow_new_fun = link_bandwidth_calc_flow_new;
    client_def->flow_set_def.flow_delete_fun = link_bandwidth_calc_flow_delete;
    client_def->flow_set_def.flow_set_client_data = lbc;

    client->link_bandwidth_calc = lbc;

    return ex;
}

#if 0
    if (args[ARG_SIMULATION].s[0] != '\0') {
	num_servers = 2;
    }
    /* Run in simulation client mode. */
    if (args[ARG_SIMULATION].s[0] != '\0') {
	client_def.type = DPCapClientTypeNS;
	client_def.type_data.ns.file_name = args[ARG_SIMULATION].s;
	client_def.type_data.ns.timers = timers;
    }
#endif

static KLException *
c_output_input(
    KLBufferedOutput *bo, void *buf, unsigned int buf_len, void *client_data)
{
    char *c = buf;

    assert((buf != NULL) && (buf_len != 0) && (client_data != NULL));
    if (c[0] == 'q') {
	KLBufferedOutputFlush(bo);
	return KLExceptionThrow(&nettimer_quit_ec);
    }
    return KLBufferedOutputFlush(bo);
}

/*
 * Initialize the update file. 
 */
static KLBufferedOutput *
c_output_update_init(
    KLEventManager *manager, KLParseOptionValue *args,
    KLBufferedOutputDef *bo_def)
{
    KLBufferedOutput *bo = NULL;

    if (args[ARG_UPDATE_FILE_NAME].s[0] == '\0') {
	return NULL;
    }

    bo_def->manager = manager;
    KLTimingSetDouble(&bo_def->timeout, args[ARG_UPDATE_INTERVAL].d);
    bo_def->flush_flags = KLBufferedOutputFlushFlagTimeout | 
	KLBufferedOutputFlushFlagForce;
    bo_def->new_client = NULL;
    bo_def->delete_client = NULL;
    bo_def->verbosity = args[ARG_VERBOSE].i;

    /* If a file name specified, write to file. */
    if (args[ARG_UPDATE_FILE_NAME].s[0] != '-') {
	bo_def->input = NULL;
	bo_def->type = KLBufferedOutputTypeFile;
	bo_def->type_data.file.file_name = args[ARG_UPDATE_FILE_NAME].s;
    }
    /* If stdout is a terminal, write to stdout. */
    else if (isatty(fileno(stdout))) {
	bo_def->type = KLBufferedOutputTypeTerminal;
	bo_def->input = c_output_input;
    }
    /* Write to stdout as if it were a file. */
    else {
	bo_def->input = NULL;
	bo_def->type = KLBufferedOutputTypeFile;
	bo_def->type_data.file.fd = fileno(stdout);
	bo_def->type_data.file.file_name = "";
    }
    KLExceptionErrorCatch(KLBufferedOutputOpen(bo_def, &bo), "");

    return bo;
}

static KLException * 
c_client_init_finished(void *client_data)
{
    Client *client;
    
    client = client_data;
    if (client->link_bandwidth_calc != NULL) {
	link_bandwidth_calc_start_sending(client->link_bandwidth_calc);
    }
    else if (client->ap != NULL) {
	active_probing_start(client->ap);
    }
    return NULL;
}

/*
 * Initialize the packet capture client.
 */
static KLException *
c_dpcap_client_init(
    Client *client, KLEventManager *manager, KLParseOptionValue *args,
    DPCapClientDef *client_def, DPCapClient **client_p)
{
    KLParseOptionServers *servers = NULL;
    KLException *ex = NULL;
    KLList *file_names = NULL;

    client_def->manager = manager;
    client_def->flow_set_def.max_flow_len = args[ARG_MAX_FLOW_LEN].i;
    client_def->flow_set_def.max_num_flows = args[ARG_MAX_NUM_FLOWS].i;
    client_def->flow_set_def.verbosity = args[ARG_VERBOSE].i;
    /* XXX What's with this? */
    client_def->flow_set_def.min_cap_len = args[ARG_DPCAP_CAP_LEN].i;

    client_def->filter_cmd = args[ARG_DPCAP_FILTER].s;
    client_def->cap_len = args[ARG_DPCAP_CAP_LEN].i;
    client_def->verbosity = args[ARG_VERBOSE].i;
    client_def->init_finished = c_client_init_finished;
    client_def->init_finished_client_data = client;

    if (args[ARG_READ_TRACE].s[0] != '\0') {
	/* Read from traces. */
	client_def->type = DPCapClientTypeTrace;
	file_names = KLStringSplitNew(args[ARG_READ_TRACE].s, " ");
	client_def->type_data.trace.file_names = file_names;
	client_def->flow_set_def.num_timers = KLListCount(
	    client_def->type_data.trace.file_names);
    }
    else {
	/* Run in live client mode. */
	servers = KLParseOptionServersParse(
	    args[ARG_DPCAP_SERVERS].s, args[ARG_DPCAP_SERVER_PORT].s);
	client_def->flow_set_def.num_timers = KLParseOptionNumServers(servers);

	client_def->type = DPCapClientTypeLive;
	client_def->type_data.live.servers = servers;
	client_def->type_data.live.interface = args[ARG_INTERFACE].s;
    }

    *client_p = NULL;
    ex = DPCapClientNew(client_def, client_p);
    if (servers != NULL) {
	KLParseOptionServersFree(servers);
    }
    if (file_names != NULL) {
	KLStringSplitFree(file_names);
    }
    return ex;
}

/*
 * Initialize the flow calc client.
 */
static KLException *
c_active_probing_init(
    Client *client, KLEventManager *manager, KLParseOptionValue *args)
{
    ActiveProbingDef ap_def;
    KLException *ex = NULL;
    int len;
    KLArray *a;

    ap_def.manager = manager;
    c_active_probe_flow_init(args, &ap_def.flow);
    /* Set local capture. */
    a = KLArrayNew();
    KLArrayStrCat(a, "localhost");
    args[ARG_DPCAP_SERVERS].s = KLArrayGetBuffer(a, &len);
    KLArrayFree(a, 0);
    /* Set special filter. */
    //a = KLArrayNew();
    //KLArrayStrCat(a, "tcp and host %s", dst_name);
    //args[ARG_DPCAP_FILTER].s = KLArrayGetBuffer(a, &len);
    //KLArrayFree(a, 0);

    KLTimingSetDouble(
	&ap_def.timeout, args[ARG_ACTIVE_PROBING_TIMEOUT].d);

    ap_def.burst_size = args[ARG_ACTIVE_PROBING_BURST_SIZE].i;
    ap_def.packet_size = args[ARG_ACTIVE_PROBING_PACKET_SIZE].i;
    ap_def.min_samples = 
	args[ARG_ACTIVE_PROBING_MIN_SAMPLES].i;
    ap_def.max_packets_sent = args[ARG_ACTIVE_PROBING_MAX_PACKETS_SENT].i;
    ap_def.max_error = args[ARG_ACTIVE_PROBING_MAX_ERROR].d;

    client->ap = active_probing_new(&ap_def);

    args[ARG_UPDATE_FILE_NAME].s = malloc(1);
    args[ARG_UPDATE_FILE_NAME].s[0] = '\0';
    return ex;
}

/*
 * Initialize the flow calc client.
 */
static KLException *
c_flow_calc_client_init(
    Client *client, KLEventManager *manager, KLParseOptionValue *args, 
    DPCapClientDef *client_def, KLBufferedOutputDef *bo_def,
    FlowCalcPacketNewFun packet_new, void *client_data)
{
    CalcDef calc_def;
    FlowCalcSetDef flow_calc_set_def;
    KLException *ex = NULL;

    /* Run in packet pair mode. */
    fill_calc_def(args, &calc_def);
    if (args[ARG_LOG_FILE_NAME].s[0] != '\0') {
	ex = c_output_log_init(
	    manager, args, args[ARG_LOG_FILE_NAME].s, &client->result_log);
	if (ex != NULL) {
	    return ex;
	}
    }
    if (args[ARG_SAMPLE_FILE_NAME].s[0] != '\0') {
	ex = c_output_log_init(
	    manager, args, args[ARG_SAMPLE_FILE_NAME].s, &client->sample_log);
	if (ex != NULL) {
	    return ex;
	}
    }
    memcpy(&flow_calc_set_def.calc_spec, &calc_def, sizeof(CalcDef));
    flow_calc_set_def.result_log = client->result_log;
    flow_calc_set_def.sample_log = client->sample_log;
    flow_calc_set_def.no_convert_addresses = 
	args[ARG_NO_CONVERT_ADDRESSES].i;
    flow_calc_set_def.qualify_host_names = 
	args[ARG_QUALIFY_HOST_NAMES].i;
    flow_calc_set_def.packet_new = packet_new;
    flow_calc_set_def.client_data = client_data;
    client->flow_calc_set = flow_calc_set_new(&flow_calc_set_def);

    bo_def->update = flow_calc_update_output;
    bo_def->client_data = client->flow_calc_set;

    client_def->flow_set_def.packet_new_fun = flow_calc_set_packet_new;
    client_def->flow_set_def.flow_new_fun = flow_calc_set_flow_new;
    client_def->flow_set_def.flow_delete_fun = flow_calc_set_flow_delete;
    client_def->flow_set_def.flow_set_client_data = client->flow_calc_set;

    return NULL;
}

/*
 * Initialize a client.
 */
KLException *
client_new(
    KLEventManager *manager, KLParseOptionValue *args, Client **client_p)
{
    Client *client;
    KLBufferedOutputDef bo_def;
    DPCapClientDef client_def;
    KLException *ex = NULL;
    FlowCalcPacketNewFun packet_new;
    void *client_data;

    client = malloc(sizeof(Client));
    client->link_bandwidth_calc = NULL;
    client->ap = NULL;
    client->flow_calc_set = NULL;
    client->sample_log = NULL;
    client->result_log = NULL;

    /* Initialize the link_bandwidth calculation. */
    if (args[ARG_ACTIVE_PROBING_ADDRESSES].s[0] != '\0') {
	if (args[ARG_ALL_LINK_BANDWIDTH].i) {
	    /* Run in link bandwidth client mode. */
	    c_lbc_init(client, manager, args, &client_def, &bo_def);
	}
	else {
	    ex = c_active_probing_init(client, manager, args);
	    if (ex != NULL) {
		return ex;
	    }
	    packet_new = active_probing_packet_new;
	    client_data = client->ap;
	    /* This is ugly. */
	    args[ARG_NO_CONVERT_ADDRESSES].i = 1;
	    ex = c_flow_calc_client_init(
		client, manager, args, &client_def, &bo_def, packet_new, 
		client_data);
	    if (ex != NULL) {
		return ex;
	    }
	}
    }
    else {
	packet_new = NULL;
	client_data = NULL;
	ex = c_flow_calc_client_init(
	    client, manager, args, &client_def, &bo_def, packet_new, 
	    client_data);
	if (ex != NULL) {
	    return ex;
	}
    } 

    /* Initialize the update file. */
    client->bo = c_output_update_init(manager, args, &bo_def);

    /* Initialize the distributed packet capture client. */
    ex = c_dpcap_client_init(
	client, manager, args, &client_def, &client->dpcap_client);

    *client_p = client;
    return ex;
}

/*
 * Delete the client.
 */
void
client_delete(Client *client)
{
    if (client->bo != NULL) {
	KLBufferedOutputFlush(client->bo);
	KLExceptionErrorCatch(KLBufferedOutputClose(client->bo), "");
    }
    if (client->result_log != NULL) {
	KLBufferedOutputFlush(client->result_log);
	KLExceptionErrorCatch(KLBufferedOutputClose(client->result_log), "");
    }
    if (client->sample_log != NULL) {
	KLBufferedOutputFlush(client->sample_log);
	KLExceptionErrorCatch(KLBufferedOutputClose(client->sample_log), "");
    }
    if (client->dpcap_client != NULL) {
	DPCapClientDelete(client->dpcap_client);
    }
    if (client->link_bandwidth_calc != NULL) {
	link_bandwidth_calc_free(client->link_bandwidth_calc);
    }
    if (client->flow_calc_set != NULL) {
	flow_calc_set_delete(client->flow_calc_set);
    }
    if (client->ap != NULL) {
	active_probing_free(client->ap);
    }
    free(client);
}

