/*
 * $Id: lbc_tailgate_packet.c,v 1.6 2001/06/28 03:16:50 laik Exp $
 *
 * Copyright (C) 1998 Kevin Lai
 *
 * This module contains code for calculating the sigma bandwidth.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <assert.h>
#include <arpa/inet.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>

#include "transport.h"
#include "lbc_tailgate_packet.h"

struct LBCTailgatePacket {
    const DPCapFlowPacket *dpcap_packet; /* The captured packet */
    struct timespec delay;               /* One way or round trip delay. */
};

struct LBCTailgatePacketGroup {
    LBCTailgatePacketGroupDef def;
    LBCTailgatePacket **packets;
    struct in_addr link_addr;
    unsigned char link_addr_set;
};

static LBCTailgatePacket *
lbctp_packet_new(void)
{
    LBCTailgatePacket *packet;

    packet = malloc(sizeof(LBCTailgatePacket));
    packet->dpcap_packet = NULL;
    KLTimingSet(&packet->delay, 0, 0);

    return packet;
}

static void
lbctp_packet_free(LBCTailgatePacket *packet)
{
    free(packet);
}

LBCTailgatePacketGroup *
lbc_tailgate_packet_group_new(LBCTailgatePacketGroupDef *def)
{
    LBCTailgatePacketGroup *pg;
    unsigned int p;

    pg = malloc(sizeof(LBCTailgatePacketGroup));
    memcpy(&pg->def, def, sizeof(LBCTailgatePacketGroupDef));
    pg->packets = malloc(sizeof(LBCTailgatePacket *) * pg->def.num_packets);
    for (p = 0; p < pg->def.num_packets; p++) {
	pg->packets[p] = lbctp_packet_new();
    }
    pg->link_addr_set = 0;
    return pg;
}

void
lbc_tailgate_packet_group_free(LBCTailgatePacketGroup *pg)
{
    unsigned int p;

    for (p = 0; p < pg->def.num_packets; p++) {
	lbctp_packet_free(pg->packets[p]);
    }
    free(pg->packets);
    free(pg);
}

int
lbc_tailgate_packet_group_match(
    LBCTailgatePacketGroup *pg, unsigned short packet_id)
{
    if ((packet_id >= pg->def.packet_id) && 
	(packet_id < (pg->def.packet_id + pg->def.num_packets))) {
	return 1;
    } 
    else if ((pg->def.packet_id + pg->def.num_packets) < pg->def.packet_id) {
	if ((packet_id >= pg->def.packet_id) ||
	    (packet_id < (pg->def.packet_id + pg->def.num_packets))) {
	    return 1;
	} 
    }
    return 0;
}

void
lbc_tailgate_packet_group_new_sample(
    LBCTailgatePacketGroup *pg, 
    const DPCapFlowPacket *dpcap_packet, unsigned short packet_id,
    const struct timespec *delay)
{
    LBCTailgatePacket *packet;
    unsigned int p;
    const DPCapFlowPacket *ack;
    const DPCapFlow *ack_flow;
    const struct in_addr *packet_src, *packet_dest;
    char old_addr[INET_ADDRSTRLEN], new_addr[INET_ADDRSTRLEN];
    int timer_src, timer_dst;

    p = packet_id - pg->def.packet_id;
    packet = pg->packets[p];
    packet->dpcap_packet = dpcap_packet;
    KLTimingCopy(&packet->delay, delay);
    if (p < pg->def.num_tailgated_packets) {
	ack = DPCapFlowPacketExactAck(packet->dpcap_packet);
	if (ack == NULL) {
	    return;
	}
	ack_flow = DPCapFlowPacketFlow(ack);
	DPCapFlowGetEndPoints(
	    ack_flow, &packet_src, &timer_src, &packet_dest, &timer_dst);
	if (pg->link_addr_set == 0) {
	    memcpy(&pg->link_addr, packet_src, sizeof(struct in_addr));
	    pg->link_addr_set = 1;
	}
	else if (memcmp(&pg->link_addr, packet_src, 
			sizeof(struct in_addr)) != 0) {
	    inet_ntop(
		AF_INET, &pg->link_addr, old_addr, sizeof(old_addr));
	    inet_ntop(
		AF_INET, packet_src, new_addr, sizeof(new_addr));
	    printf("Address of node %d changed from %16s to %16s in group.\n",
		   pg->def.ttl, old_addr, new_addr);
	    
	}
    }
}

const struct in_addr *
lbc_tailgate_packet_group_get_addr(LBCTailgatePacketGroup *pg)
{
    if (pg->link_addr_set) {
	return &pg->link_addr;
    }
    else {
	return NULL;
    } 
}

unsigned int 
lbc_tailgate_packet_group_get_ttl(LBCTailgatePacketGroup *pg)
{
    return pg->def.ttl;
}

void
lbc_tailgate_packet_group_get_packet_size(
    LBCTailgatePacketGroup *pg, unsigned int *ack_size, 
    unsigned int *tailgater_size)
{
    const DPCapFlowPacket *ack;
    LBCTailgatePacket *packet;
    
    packet = pg->packets[pg->def.num_packets - 1];
    ack = DPCapFlowPacketExactAck(packet->dpcap_packet);
    *ack_size = DPCapFlowPacketGetSize(ack);
    *tailgater_size = pg->def.tailgater_size;
}

int
lbc_tailgate_packet_group_check_finished(LBCTailgatePacketGroup *pg)
{
    unsigned int p, finished;
    LBCTailgatePacket *packet;
    
    finished = 1;
    for (p = 0; p < pg->def.num_packets; p++) {
	packet = pg->packets[p];
	if (KLTimingGetDouble(&packet->delay) == 0.0) {
	    finished = 0;
	    break;
	}
    }
    return finished;
}

void
lbc_tailgate_packet_group_get_delay(
    LBCTailgatePacketGroup *pg, struct timespec *delay)
{
    struct timespec *large_timing, *small_timing;
    LBCTailgatePacket *first, *last;

    first = pg->packets[0];
    last = pg->packets[pg->def.num_packets - 1];

    /* Compute the difference in transmission delay. */
    large_timing = DPCapFlowPacketTimings(first->dpcap_packet);
    small_timing = DPCapFlowPacketTimings(last->dpcap_packet);
    KLTimingSubtract(delay, &small_timing[0], &large_timing[0]);

    /* Compute best result. */
    /* Use the rtt of the last sent packet. */ 
    KLTimingAdd(delay, &last->delay, delay);
    KLTimingNormalize(delay);
}

const DPCapFlowPacket *
lbc_tailgate_packet_group_get_packet(LBCTailgatePacketGroup *pg)
{
    return pg->packets[pg->def.num_packets - 1]->dpcap_packet;
}

