/*
 * $Id: link_bandwidth_calc.c,v 1.10 2001/07/19 21:52:20 laik Exp $
 *
 * Copyright (C) 1998 Kevin Lai
 *
 * This module contains code for calculating the sigma bandwidth.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>

#include "kl_array.h"

#include "lbc_sigma.h"
#include "lbc_tailgate.h"
#include "link_bandwidth_calc.h"

typedef enum {
    lbc_packet_type_none,
    lbc_packet_type_sigma,
    lbc_packet_type_tailgate,
} lbc_packet_type;

struct LinkBandwidthCalc {
    LinkBandwidthCalcDef def;           /* Definitions. */
    LBCSigma *sigma;                    /* Sigma state. */
    LBCTailgate *tailgate;              /* Tailgate state. */
    KLEventDef timer_def;
    KLEventDefRegistration *timer_reg;    /* For the timer. */
    lbc_packet_type last_packet_sent;

    unsigned int sigma_packets_sent;       /* # packets sent this phase. */
    unsigned int sigma_packets_received;   /* # packets received this phase. */
    unsigned int sigma_packet_timeouts;    /* timeouts. */
    unsigned int tailgate_packets_sent;    /* # packets sent this phase. */
    unsigned int tailgate_packets_received;/* # packets received this phase. */
    unsigned int tailgate_packet_timeouts; /* timeouts. */
};

KLExceptionClass lbc_ec_below_max_tailgate_error[] = {
    {"Below max tailgate error."}
};

KLExceptionClass lbc_ec_exceeded_tailgate_send_limit[] = {
    {"Exceeded tailgate sent packets limit."}
};

static void
lbc_tailgate_log_output_links(LinkBandwidthCalc *lbc)
{
    KLArray *array;
    int len;
    LBCSigmaStatistics sigma_stats;

    if (lbc->def.output_log == NULL) {
	return;
    }
    lbc_sigma_get_statistics(lbc->sigma, &sigma_stats, NULL);
    array = KLArrayNew();
    lbc_links_get_string(
	lbc->tailgate, sigma_stats.slope, sigma_stats.intercept, array);
    KLBufferedOutputPrintf(
	lbc->def.output_log, NULL, "%s\n", 
	(char *) KLArrayGetBuffer(array, &len));
    KLArrayFree(array, TRUE);
}

static void
lbc_send_tailgate_packets(LinkBandwidthCalc *lbc)
{
    lbc->tailgate_packets_sent += lbc_tailgate_send_packet(lbc->tailgate);
    lbc->last_packet_sent = lbc_packet_type_tailgate;
    KLEventDefUnregister(lbc->def.manager, lbc->timer_reg);
    lbc->timer_reg = KLEventDefRegister(lbc->def.manager, &lbc->timer_def);
}
 
static void
lbc_send_sigma_packets(LinkBandwidthCalc *lbc)
{
    lbc->sigma_packets_sent += lbc_sigma_send_packet(lbc->sigma);
    lbc->last_packet_sent = lbc_packet_type_sigma;
    KLEventDefUnregister(lbc->def.manager, lbc->timer_reg);
    lbc->timer_reg = KLEventDefRegister(lbc->def.manager, &lbc->timer_def);
}
 
/*
 * Print out information about sigma phase.
 */
static void
lbc_sigma_finish(LinkBandwidthCalc *lbc)
{
    KLArray *array;
    int len;
    LBCSigmaStatistics sigma_stats;

    if (lbc->def.output_log != NULL) {
	array = KLArrayNew();
	lbc_sigma_get_statistics(lbc->sigma, &sigma_stats, array);
	KLBufferedOutputPrintf(
	    lbc->def.output_log, NULL, "%s\n", 
	    (char *) KLArrayGetBuffer(array, &len));
	KLArrayFree(array, TRUE);

	KLBufferedOutputPrintf(
	    lbc->def.output_log, NULL, "Sigma phase statistics:\n");
	KLBufferedOutputPrintf(
	    lbc->def.output_log, NULL, "  packets sent: %d \n", 
	    lbc->sigma_packets_sent);
	KLBufferedOutputPrintf(
	    lbc->def.output_log, NULL, "  packets received: %d\n", 
	    lbc->sigma_packets_received);
	KLBufferedOutputPrintf(
	    lbc->def.output_log, NULL, "  timeouts: %d\n", 
	    lbc->sigma_packet_timeouts);
    }
}

/*
 * This function handles the expiration of the sending timer event.
 * The timer prevents us from waiting too long before sending the
 * next packet.
 */
static KLException *
lbc_timer_event_handler(const KLEvent *event)
{
    LinkBandwidthCalc *lbc = event->def->client_data;

    if (lbc->last_packet_sent == lbc_packet_type_sigma) {
	/* printf("A sigma packet lost.\n"); */
	lbc->sigma_packet_timeouts++;
	lbc_send_sigma_packets(lbc);
    }
    else if (lbc->last_packet_sent == lbc_packet_type_tailgate) {
	/* printf("A tailgate packet lost.\n"); */
	lbc->tailgate_packet_timeouts++;
	lbc_send_tailgate_packets(lbc);
    }
    return NULL;
}

/*
 * Determine if we have the full timing information about a packet and
 * compute delay.
 */
static unsigned int
lbc_check_packet(
    DPCapFlow *flow, const DPCapFlowPacket *packet, int ack,
    struct timespec *delay)
{
    const DPCapFlowPacket *ack_packet;
    struct timespec *timings, *ack_timings;
    DPCapFlowDirection direction;
    const DPCapFlowSet *flow_set;
    int num_timers;

    timings = DPCapFlowPacketTimings(packet);
    direction = DPCapFlowGetDirection(flow);
    flow_set = DPCapFlowGetSet(flow);
    num_timers = DPCapFlowSetNumTimers(flow_set);
    /* Compute round trip or one-way delay. */
    KLTimingSet(delay, 0, 0);
    if (ack) {
	ack_packet = DPCapFlowPacketExactAck(packet);
	ack_timings = DPCapFlowPacketTimings(ack_packet);

	/*
	 * Get delay when we get an ack for a packet that
	 * we sent.
	 */
	if ((direction == DPCapFlowDirectionForward) && 
	    ((KLTimingGetSeconds(&timings[0]) != 0) &&
	     (KLTimingGetNS(&timings[0]) != 0)) && 
	    ((KLTimingGetSeconds(&ack_timings[0]) != 0) &&
	     (KLTimingGetNS(&ack_timings[0]) != 0))) {
	    KLTimingSubtract(delay, &ack_timings[0], &timings[0]);
	    KLTimingNormalize(delay);
	    return 2;
	}
    }
    else {
	/*
	 * Get delay when we have the source and destination timings
	 * of a packet.
	 */
	if ((direction == DPCapFlowDirectionForward) && 
	    ((KLTimingGetSeconds(&timings[0]) != 0) &&
	     (KLTimingGetNS(&timings[0]) != 0)) && 
	    ((KLTimingGetSeconds(&timings[num_timers - 1]) != 0) &&
	     (KLTimingGetNS(&timings[num_timers - 1]) != 0))) {
	    KLTimingSubtract(delay, &timings[num_timers - 1], &timings[0]);
	    KLTimingNormalize(delay);
	    return 1;
	}
    }
    return 0;
}

/*
 * Allocate and initialize a new link bandwidth calculation state.
 */
LinkBandwidthCalc *
link_bandwidth_calc_new(LinkBandwidthCalcDef *def)
{
    LBCSigmaDef sigma_def;
    LBCTailgateDef tailgate_def;
    LinkBandwidthCalc *lbc;

    lbc = malloc(sizeof(LinkBandwidthCalc));
    memcpy(&lbc->def, def, sizeof(LinkBandwidthCalcDef));

    /* Create sigma state. */
    memcpy(&sigma_def.flow, &lbc->def.flow, sizeof(TransportFlow));
    sigma_def.num_packet_sizes = lbc->def.sigma_min_packets / 32;
    sigma_def.receive_per_size = 32;
    lbc->sigma = lbc_sigma_new(&sigma_def);

    /* Create tailgate state. */
    memcpy(&tailgate_def.flow, &lbc->def.flow, sizeof(TransportFlow));
    ((struct sockaddr_in *) &tailgate_def.flow.source.sock_addr)->sin_port =
	htons(ntohs(((struct sockaddr_in *) 
		     &tailgate_def.flow.source.sock_addr)->sin_port) + 1);
    tailgate_def.min_packets = lbc->def.tailgate_min_packets;
    tailgate_def.max_bootstrap_sample_size = 100;
    tailgate_def.bootstrap_iterations = 100;
    lbc->tailgate = lbc_tailgate_new(&tailgate_def);

    lbc->sigma_packets_sent = 0;
    lbc->sigma_packets_received = 0;
    lbc->sigma_packet_timeouts = 0;
    lbc->tailgate_packets_sent = 0;
    lbc->tailgate_packets_received = 0;
    lbc->tailgate_packet_timeouts = 0;
    lbc->last_packet_sent = lbc_packet_type_none;

    /* Register sending function on a timer. */
    /* XXX Should update timer. */
    lbc->timer_def.type = KLEventTypeTimer;
    lbc->timer_def.client_data_delete_fun = NULL;
    lbc->timer_def.client_data = lbc;
    lbc->timer_def.handler_fun = lbc_timer_event_handler;
    KLTimingCopy(&lbc->timer_def.type_data.timer_interval,
	       &lbc->def.max_sending_interval);
    lbc->timer_reg = KLEventDefRegister(lbc->def.manager, &lbc->timer_def);
    return lbc;
}

/*
 * Free resources associated with a link bandwidth calculation state.
 */
void
link_bandwidth_calc_free(LinkBandwidthCalc *lbc)
{
    KLEventDefUnregister(lbc->def.manager, lbc->timer_reg);
    lbc_sigma_free(lbc->sigma);
    lbc_tailgate_free(lbc->tailgate);
    free(lbc);
}

/*
 * 
 */
void
link_bandwidth_calc_start_sending(LinkBandwidthCalc *lbc)
{
    lbc_sigma_send_packet(lbc->sigma);
    lbc->sigma_packets_sent++;
    lbc->last_packet_sent = lbc_packet_type_sigma;
    KLEventDefUnregister(lbc->def.manager, lbc->timer_reg);
    lbc->timer_reg = KLEventDefRegister(lbc->def.manager, &lbc->timer_def);
}

/*
 * Called when its time to write to the update file. 
 */
KLException *
link_bandwidth_calc_update_output(
    KLBufferedOutput *bo, void *client_data)
{
    LinkBandwidthCalc *lbc = client_data;
    LBCSigmaStatistics sigma_stats;
    KLArray *array;
    int len;
    KLException *ex = NULL;

    ex = KLExceptionPass(NULL, KLBufferedOutputPrintf(
	bo, NULL, "Cmds: (q)uit, any other key to update\n"));
    if (ex != NULL) {
	return ex;
    }

   /* Sigma info. */
    ex = KLExceptionPass(NULL, KLBufferedOutputPrintf(
	bo, NULL, "Sigma:    sent:%14d  received:%11d  timeouts:%13d\n",
	lbc->sigma_packets_sent, lbc->sigma_packets_received, 
	lbc->sigma_packet_timeouts));
    if (ex != NULL) {
	return ex;
    }

    lbc_sigma_get_statistics(lbc->sigma, &sigma_stats, NULL);
    ex = KLExceptionPass(NULL, KLBufferedOutputPrintf(
	bo, NULL, "          bw:%16.3f  rtt:%16.12f  conf:%17.9f\n\n",
	1 / sigma_stats.slope, sigma_stats.intercept,
	sigma_stats.confidence));
    if (ex != NULL) {
	return ex;
    }

    /* Tailgate info. */
    ex = KLExceptionPass(NULL, KLBufferedOutputPrintf(
	bo, NULL, "Tailgate: sent:%14d  received:%11d  timeouts:%13d\n\n",
	lbc->tailgate_packets_sent, lbc->tailgate_packets_received, 
	lbc->tailgate_packet_timeouts));
    if (ex != NULL) {
	return ex;
    }

    array = KLArrayNew();
    lbc_links_get_string(
	lbc->tailgate, sigma_stats.slope, sigma_stats.intercept, array);
    if (KLArrayGetLen(array) != 0) {
	ex = KLExceptionPass(NULL, KLBufferedOutputPrintf(
	    bo, NULL, "%s", (char *) KLArrayGetBuffer(array, &len)));
    }
    KLArrayFree(array, TRUE);

    return ex;
}

/*
 * Called when new information about a captured packet arrives.
 */
KLException *
link_bandwidth_calc_packet_new(
    DPCapFlow *flow, const KLListNode *cur_node, int timer_index,
    int ack, void *flow_set_client_data)
{
    LinkBandwidthCalc *lbc = flow_set_client_data;
    struct timespec delay;
    const DPCapFlowPacket *packet;
    unsigned int timing_type, num_timers;
    LBCSigmaStatistics sigma_stats;
    KLException *ex = NULL;
    double max_error;

    packet = KLListGetItem(cur_node);
    timing_type = lbc_check_packet(flow, packet, ack, &delay);
    num_timers = DPCapFlowSetNumTimers(DPCapFlowGetSet(flow));
    if ((num_timers == 1) && (timing_type == 1)) {
	timing_type = 0;
    }

    /*
     * Call both the sigma and tailgate handlers and let them sort 
     * out whose packet it is.
     */
    if (lbc_sigma_sample_received(lbc->sigma, packet, &delay, timing_type)) {
	lbc_sigma_get_statistics(lbc->sigma, &sigma_stats, NULL);
	if (lbc->def.output_log != NULL) {
	    KLBufferedOutputPrintf(
		lbc->def.output_log, NULL, 
		"BW: %16.4f L: %16.12f C: %16.10f\n",
		1 / sigma_stats.slope, sigma_stats.intercept,
		sigma_stats.confidence);
	}
	lbc->sigma_packets_received++;
	/* 
	 * Send packets if we're below min packets or below max packets
	 * and below min confidence. 
	 */
	if ((lbc->sigma_packets_received < lbc->def.sigma_min_packets) ||
	    ((lbc->sigma_packets_received <= lbc->def.sigma_max_packets) &&
	     (sigma_stats.confidence < lbc->def.sigma_min_confidence))) {
	    lbc_send_sigma_packets(lbc);
	}
	else {
	    lbc_sigma_finish(lbc);
	    // printf("Starting tailgating.\n");
	    lbc_send_tailgate_packets(lbc);
	}
    }
    else if (lbc_tailgate_sample_received(
	lbc->tailgate, packet, &delay, timing_type)) {
	lbc->tailgate_packets_received += 2;
	lbc_tailgate_log_output_links(lbc);
	
	lbc_tailgate_get_max_error(lbc->tailgate, &max_error);
	/* Check if we can continue tailgating. */
	if ((lbc->tailgate_packets_sent < lbc->def.tailgate_min_packets) ||
	    ((max_error > lbc->def.tailgate_max_error) &&
	     (lbc->tailgate_packets_sent < lbc->def.tailgate_max_packets))) {
	    lbc_send_tailgate_packets(lbc);
	}
	else {
	    if (max_error < lbc->def.tailgate_max_error) {
		ex = KLExceptionThrow(lbc_ec_below_max_tailgate_error);
	    }
	    else {
		ex = KLExceptionThrow(lbc_ec_exceeded_tailgate_send_limit);
	    }
	}
    }
    return ex;
}

void
link_bandwidth_calc_flow_new(
    DPCapFlow *flow, void *flow_set_client_data)
{
    assert(flow != NULL);
    assert(flow_set_client_data != NULL);
}

void
link_bandwidth_calc_flow_delete(
    DPCapFlow *flow, void *flow_set_client_data)
{
    assert(flow != NULL);
    assert(flow_set_client_data != NULL);
}
