/*
 * $Id: nettimer.c,v 1.37 2001/07/24 23:29:16 laik Exp $
 * 
 * This is a utility to measure the bandwidth of a path in a 
 * packet-switched network like the Internet. 
 *
 * Live passive client mode:
 *     nettimer -t ["server1 server2"]
 *
 * Trace passive client mode:
 *     nettimer -r "trace1 ..."
 *
 * Live passive server mode:
 *     nettimer -s 
 *
 * Trace passive server mode:
 *     nettimer -s -r trace1
 *
 * Copyright (C) 1998 Kevin Lai
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <assert.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "kl_array.h"
#include "kl_event.h"
#include "kl_timing.h"
#include "kl_version.h"

#include "dpcap_server.h"
#include "dpcap_client.h"
#include "dpcap.h"

#include <config.h>

#include "active_probing.h"
#include "calc.h"
#include "client.h"
#include "transport.h"

#include "nettimer.h"

/* Version Information */
static KLVersion  n_version = {
    BUILD_VERSION_MAJOR,
    BUILD_VERSION_MINOR,
    BUILD_VERSION_TERTIARY,
    __DATE__,
    __TIME__,
    __VERSION__,
};

KLExceptionClass nettimer_quit_ec = {"nettimer quits"};

static KLExceptionClass n_caught_sigint_ec[] = {
    {"Caught SIGINT."}
};

static KLExceptionClass n_bad_args_ec[] = {
    {"Bad command line arguments: %s."}
};

/* These are the mappings from command line strings to scalar values. */
static KLParseOptionScalarPair calc_pp_nodes_scalars[] = {
    {"SBPP", (void *) (1 << CalcMetricSBBW)},
    {"RBPP", (void *) (1 << CalcMetricRBBW)},
    {"ROPP", (void *) (1 << CalcMetricROBW)},
    {"auto", (void *) ((1 << CalcMetricSBBW) | (1 << CalcMetricRBBW) | 
		       (1 << CalcMetricROBW))},
    {"", NULL},
};

static KLParseOptionScalarPair calc_kernel_func_scalars[] = {
    {"uniform", (void *) KLStatisticsKernelFunctionUniform},
    {"triangular", (void *) KLStatisticsKernelFunctionTriangular},
    {"", NULL},
};

static KLParseOptionScalarPair n_active_probing_protocol[] = {
    {"udp", (void *) IPPROTO_UDP},
    {"tcp", (void *) IPPROTO_TCP},
    {"icmp", (void *) IPPROTO_ICMP},
    {"", NULL},
};

/* These are the specifications for command line arguments. */
static KLParseOptionSpec nettimer_option_specs[] = {
    /* General options */
    {"event_log",           '\0', ARG_EVENT_LOG,        PO_TYPE_STRING,
     "Write events out to this log file.",
     {s: ""}, {s: "event.log"},   NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"verbose",     '\0', ARG_VERBOSE,   PO_TYPE_INT,
     "Increase the verbosity of the output.",
     {i: 0}, {i: 1},   NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"version",     'v', ARG_VERSION,   PO_TYPE_INT,
     "Print the version of this program.",
     {i: 0}, {i: 1},   NULL, PO_NO_ARG, PO_NOT_REQUIRED},

    /* Determine if we are a client or a server. */
    {"run_dpcap_server",     's', ARG_RUN_DPCAP_SERVER,   PO_TYPE_INT,
     "Run as a distributed packet capture server.",
     {i: 0}, {i: 1},   NULL, PO_NO_ARG, PO_NOT_REQUIRED},

    /* Apply to both client and server. */
    {"interface",      'i', ARG_INTERFACE,     PO_TYPE_STRING,
     "Capture packets from this network interface.",
     {s: ""}, {s: ""}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},
     
    {"dpcap_filter",  'e', ARG_DPCAP_FILTER,     PO_TYPE_STRING,
     "Filter captured packets using this expression.",
     {s: ""}, {s: ""}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"dpcap_cap_len",  '\0', ARG_DPCAP_CAP_LEN,      PO_TYPE_INT,
     "Capture this many bytes of each packet.",
     {i: 60}, {i: 0}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    /* dpcap server misc. options */
    {"dpcap_server_port",   'p', ARG_DPCAP_SERVER_PORT,      PO_TYPE_STRING,
     "The dpcap server listens on this port for client connections.",
     {s: "9090"}, {s: ""}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    /* dpcap server capture arguments */
    {"dpcap_server_packet_buffer", '\0', ARG_DPCAP_SERVER_PACKET_BUFFER,
     PO_TYPE_INT,
     "Buffer this many bytes before sending to clients.",
     {i: 8000}, {i: 0}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"dpcap_server_send_timeout", '\0', ARG_DPCAP_SERVER_SEND_TIMEOUT,
     PO_TYPE_DOUBLE,
     "Wait at most this many seconds before sending to the clients.",
     {d: 1.0}, {d: 0.0}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    /* Client misc options */
    {"dpcap_servers",        't', ARG_DPCAP_SERVERS,     PO_TYPE_STRING,
     "Use or connect to the specified distributed packet capture servers.",
     {s: ""}, {s: "localhost"},   NULL, PO_REQUIRED_ARG,
     PO_NOT_REQUIRED},

    {"read_trace",          'r', ARG_READ_TRACE, PO_TYPE_STRING,
     "Read traces from the specified \"file1 file2 ...\".",
     {s: ""}, {s: ""}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"sample_file_name",       '\0', ARG_SAMPLE_FILE_NAME,     PO_TYPE_STRING,
     "Write raw samples to this file.",
     {s: ""}, {s: ""}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"log_file_name",         'w', ARG_LOG_FILE_NAME,      PO_TYPE_STRING,
     "Write results to this file. Use \"-\" to write to stdout.",
     {s: ""}, {s: ""},   NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"update_file_name",    'u', ARG_UPDATE_FILE_NAME,  PO_TYPE_STRING,
     "Update this file continuously. Use \"-\" to write to stdout.",
     {s: "-"}, {s: ""},  NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"update_interval",    '\0', ARG_UPDATE_INTERVAL,  PO_TYPE_DOUBLE,
     "Update the update file every n seconds.",
     {d: 1.0}, {d: 1.0},   NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"no_convert_addresses", 'n', ARG_NO_CONVERT_ADDRESSES, PO_TYPE_INT,
     "Don't convert network addresses to host names.",
     {i: 0}, {i: 1},   NULL, PO_NO_ARG, PO_NOT_REQUIRED},

    {"qualify_host_names",  'N', ARG_QUALIFY_HOST_NAMES,  PO_TYPE_INT,
     "Qualify host names with domain names.",
     {i: 0}, {i: 1},   NULL, PO_NO_ARG, PO_NOT_REQUIRED},

    /* Client capture options */
    {"maximum_number_of_flows",   '\0', ARG_MAX_NUM_FLOWS,   PO_TYPE_INT,
     "Keep track of at most this many flows.",
     {i: 100}, {i: 0}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"maximum_flow_length",   '\0', ARG_MAX_FLOW_LEN,   PO_TYPE_INT,
     "Keep track of at most this many packets per flow.",
     {i: 10000}, {i: 0}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    /* Client calculation options */
    {"calc_kernel_weight",      '\0', ARG_CALC_KERNEL_WEIGHT,   PO_TYPE_DOUBLE,
     "Weight of the kernel component in packet pair filtering.",
     {d: 0.4}, {i: 0},   NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"calc_pbw_weight",      '\0', ARG_CALC_PBW_WEIGHT,   PO_TYPE_DOUBLE,
     "Weight of the potential bandwidth component in packet pair filtering.",
     {d: 0.325}, {i: 0},   NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"calc_mbw_weight",    '\0', ARG_CALC_MBW_WEIGHT,   PO_TYPE_DOUBLE,
     "Weight of the measured bandwidth component in packet pair filtering.",
     {d: 0.275}, {i: 0},   NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"calc_min_bunch_size",       '\0', ARG_CALC_MIN_BUNCH_SIZE,  PO_TYPE_INT,
     "Use packet bunch sizes at least this large.",
     {i: 2}, {i: 0},   NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"calc_max_bunch_size",       '\0', ARG_CALC_MAX_BUNCH_SIZE,  PO_TYPE_INT,
     "Use packet bunch sizes no larger than this.",
     {i: 6}, {i: 0},   NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"calc_restrict_packet_pair", '\0', ARG_CALC_RESTRICT_PACKET_PAIR,
     PO_TYPE_SCALAR,
     "Only use 1 packet pair algorithm: sender-, receiver- or, receiver-only.",
     {a: (void *) ((1 << CalcMetricROBW) | (1 << CalcMetricRBBW) |
      (1 << CalcMetricSBBW))},
     {a: (void *) 0}, calc_pp_nodes_scalars, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"calc_min_delta_rtt",    '\0', ARG_CALC_MIN_DELTA_RTT,    PO_TYPE_DOUBLE,
     "Use this as the min error allowed in the RTT.",
     {d: 0.0}, {d: 0.0},   NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"calc_max_delta_rtt",    '\0', ARG_CALC_MAX_DELTA_RTT,    PO_TYPE_DOUBLE,
     "Use this as the max error allowed in the RTT.",
     {d: 0.1}, {d: 0.0},   NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"calc_window_size",       '\0', ARG_CALC_WINDOW_SIZE,    PO_TYPE_INT,
     "Use no more than this number of packets into the past.",
     {i: 128}, {i: 0},   NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"calc_no_bw_ratio",      '\0', ARG_CALC_NO_BW_RATIO,    PO_TYPE_DOUBLE,
     "Use this as the ratio of good samples below which we cannot find a bw.",
     {d: 0.05}, {i: 0},   NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"calc_kernel_bound_ratio", '\0', ARG_CALC_KERNEL_BOUND_RATIO,
     PO_TYPE_DOUBLE,
     "Set the kernel width to be (1 + this ratio) * x.",
     {d: 0.05}, {d: 0.0},   NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"calc_kernel_resolution", '\0', ARG_CALC_KERNEL_RESOLUTION,
     PO_TYPE_DOUBLE,
     "Aggregate kernels at this resolution.",
     {d: 500.0}, {d: 0.0},   NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"calc_kernel_function", '\0', ARG_CALC_KERNEL_FUNCTION,
     PO_TYPE_SCALAR,
     "Use this function to compute the kernel value.",
     {a: (void *) KLStatisticsKernelFunctionTriangular},
     {a: (void *) KLStatisticsKernelFunctionTriangular}, 
     calc_kernel_func_scalars, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    /* active probing options */
    {"active_probing_addresses", 'a', 
     ARG_ACTIVE_PROBING_ADDRESSES, PO_TYPE_STRING,
     "Actively probe \"[src_addr[:src_service] ]dst_addr[:dst_service]\".",
     {s: ""}, {s: ""}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"active_probing_protocol", '\0', 
     ARG_ACTIVE_PROBING_PROTOCOL, PO_TYPE_SCALAR,
     "Actively probe using this protocol.",
     {a: (void *) IPPROTO_TCP}, {a: (void *) IPPROTO_TCP},
     n_active_probing_protocol, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"active_probing_burst_size", '\0', 
     ARG_ACTIVE_PROBING_BURST_SIZE, PO_TYPE_INT,
     "Actively probe using bursts of this size.",
     {i: 4}, {i: 4}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"active_probing_packet_size", '\0', 
     ARG_ACTIVE_PROBING_PACKET_SIZE, PO_TYPE_INT,
     "Actively probe using packets of this size.",
     {i: 500}, {i: 1500}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"active_probing_min_samples", '\0', 
     ARG_ACTIVE_PROBING_MIN_SAMPLES, PO_TYPE_INT,
     "Actively probe until computing this many valid samples.",
     {i: 4}, {i: 4}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"active_probing_max_packets_sent", '\0', 
     ARG_ACTIVE_PROBING_MAX_PACKETS_SENT, PO_TYPE_INT,
     "Actively probe sending no more than this many packets.",
     {i: 16}, {i: 16}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"active_probing_max_error", '\0', 
     ARG_ACTIVE_PROBING_MAX_ERROR, PO_TYPE_DOUBLE,
     "Actively probe until reaching this error.",
     {d: 0.20}, {d: 0.20}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"active_probing_timeout", '\0', 
     ARG_ACTIVE_PROBING_TIMEOUT,  PO_TYPE_DOUBLE,
     "Actively probe for no longer than this many seconds.",
     {d: 5.0}, {d: 0.0}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    /* Link bandwidth mode options */
    {"all_link_bw", 'l', ARG_ALL_LINK_BANDWIDTH, PO_TYPE_INT,
     "Measure all link bandwidths along the path",
     {i: 0}, {i: 1}, NULL, PO_NO_ARG, PO_NOT_REQUIRED},

    {"all_link_bw_sigma_min_packets", '\0', 
     ARG_ALL_LINK_BANDWIDTH_SIGMA_MIN_PACKETS, PO_TYPE_INT,
     "Send at least this many sigma packets before accepting estimate.",
     {i: 128}, {i: 128}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"all_link_bw_sigma_max_packets", '\0', 
     ARG_ALL_LINK_BANDWIDTH_SIGMA_MAX_PACKETS, PO_TYPE_INT,
     "Send at most this many sigma packets.",
     {i: 16384}, {i: 16384}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"all_link_bw_sigma_min_confidence", '\0',
     ARG_ALL_LINK_BANDWIDTH_SIGMA_MIN_CONFIDENCE,  PO_TYPE_DOUBLE,
     "Continue until sigma confidence is at least this value.",
     {d: 0.9995}, {d: 0.9995}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"all_link_bw_tailgate_min_packets", '\0', 
     ARG_ALL_LINK_BANDWIDTH_TAILGATE_MIN_PACKETS, PO_TYPE_INT,
     "Send at least this many tailgate packets before accepting estimate.",
     {i: 128}, {i: 128}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"all_link_bw_tailgate_max_packets", '\0',
     ARG_ALL_LINK_BANDWIDTH_TAILGATE_MAX_PACKETS, PO_TYPE_INT,
     "Send at most this many tailgate packets.",
     {i: 16384}, {i: 16384}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"all_link_bw_tailgate_max_error", '\0',
     ARG_ALL_LINK_BANDWIDTH_TAILGATE_MAX_ERROR,  PO_TYPE_DOUBLE,
     "Continue until tailgate below this value.",
     {d: 0.0075}, {d: 0.0075}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"all_link_bw_max_sending_interval", '\0',
     ARG_ALL_LINK_BANDWIDTH_MAX_SENDING_INTERVAL,  PO_TYPE_DOUBLE,
     "Send packets no slower than at this interval.",
     {d: 1.0}, {d: 1.0}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {"all_link_bw_timeout", '\0', 
     ARG_ALL_LINK_BANDWIDTH_TIMEOUT,  PO_TYPE_DOUBLE,
     "Wait at most this many seconds before stopping calculation.",
     {d: 60.0}, {d: 0.0}, NULL, PO_REQUIRED_ARG, PO_NOT_REQUIRED},

    {NULL,                 '\0', ARG_LAST,                 PO_TYPE_LAST,
     "",
     {i: -1}, {i: -1},     NULL, PO_NO_ARG, PO_NOT_REQUIRED}
};

/*
 * This function parses the command line and verifies the semantic
 * correctness of the specified values.
 */
static KLException *
n_parse_command_line(
    int argc, char *argv[], KLParseOptionValue **args_p, int *run_server)
{
    KLParseOptionValue *args;
    int remaining;
    KLException *ex = NULL;
    KLArray *array;
    size_t len;

    /* Parse the regular options. */
    ex = KLParseOptionParse(
	argc, argv, nettimer_option_specs, &args, &remaining);
    if (ex != NULL) {
	goto done;
    }

    /* Check that the bunch sizes are correct. */
    if ((args[ARG_CALC_MAX_BUNCH_SIZE].i < 2) ||
	(args[ARG_CALC_MAX_BUNCH_SIZE].i > CALC_BUNCH_SIZE_MAX)) {
	ex = KLExceptionThrow(
	    n_bad_args_ec,
	    "The max packet bunch size (%ld) must be in [2,%d]",
	    args[ARG_CALC_MAX_BUNCH_SIZE].i, CALC_BUNCH_SIZE_MAX);
	goto done;
    }
    if ((args[ARG_CALC_MIN_BUNCH_SIZE].i < 2) ||
	(args[ARG_CALC_MIN_BUNCH_SIZE].i > CALC_BUNCH_SIZE_MAX)) {
	ex = KLExceptionThrow(
	    n_bad_args_ec,
	    "The min packet bunch size (%ld) must be in [2,%d]",
	    args[ARG_CALC_MIN_BUNCH_SIZE].i, CALC_BUNCH_SIZE_MAX);
	goto done;
    }
    if (args[ARG_CALC_MIN_BUNCH_SIZE].i > args[ARG_CALC_MAX_BUNCH_SIZE].i) {
	ex = KLExceptionThrow(
	    n_bad_args_ec,
	    "The min packet bunch size (%ld) must be in less than the max packet bunch size(%ld)", 
	    args[ARG_CALC_MIN_BUNCH_SIZE].i,
	    args[ARG_CALC_MAX_BUNCH_SIZE].i);
	goto done;
    }

    /*
     * No point in specifying an interface and reading from traces.
     */
    if ((args[ARG_INTERFACE].s[0] != '\0') &&
	(args[ARG_READ_TRACE].s[0] != '\0')) {
	ex = KLExceptionThrow(
	    n_bad_args_ec,
	    "Can't both capture from an interface and read trace file(s)");
	goto done;
    }

    /*
     * Can't specify a window larger than the maximum flow length.
     */
    if (args[ARG_CALC_WINDOW_SIZE].i > args[ARG_MAX_FLOW_LEN].i) {
	ex = KLExceptionThrow(
	    n_bad_args_ec,
	    "The window size (%d) is larger than the max flow length (%d)",
	    args[ARG_CALC_WINDOW_SIZE].i, args[ARG_MAX_FLOW_LEN].i);
	goto done;
    }

    if (args[ARG_VERSION].i) {
	array = KLArrayNew();

	KLArrayStrCat(array, "nettimer: ");
	KLVersionStrCat(&n_version, array);
	KLArrayStrCat(array, "\n");

	KLArrayStrCat(array, "libdpcap: ");
	KLVersionStrCat(DPCapVersionGet(), array);
	KLArrayStrCat(array, "\n");

	KLArrayStrCat(array, "libkl: ");
	KLVersionStrCat(KLVersionGet(), array);
	KLArrayStrCat(array, "\n");
	printf("%s", (char *) KLArrayGetBuffer(array, &len));

	KLArrayFree(array, TRUE);
	goto done;
    }

    /* Run as dpcap server. */
    if (args[ARG_RUN_DPCAP_SERVER].i == TRUE) {
	*run_server = 1;
    }
    /* Run as a client. */
    else if ((args[ARG_DPCAP_SERVERS].s[0] != '\0') ||
	     (args[ARG_READ_TRACE].s[0] != '\0') ||
	     (args[ARG_ACTIVE_PROBING_ADDRESSES].s[0] != '\0')) {
	*run_server = 0;
    }
    /* Didn't specify running as client or server. */
    else {
	ex = KLExceptionThrow(
	    n_bad_args_ec,
	    "Run as a libdpcap server or specify measurement sources");
	goto done;
    }

    argc -= remaining;
    argv += remaining;

 done:
    *args_p = args;
    return ex;
}

const char *
getprogramname()
{
    return KLUtilProgramName;
}


static KLException *
n_server_init_finished(void *client_data)
{
    assert(client_data == NULL);

    /* Make user own process after packet capture socket has been opened. */
    setuid(getuid());
    return NULL;
}

static KLException *
n_dpcap_server_init(
    KLEventManager *manager, KLParseOptionValue *args, DPCapServer **server_p)
{
    DPCapServerDef def;
    KLException *ex = NULL;

    def.manager = manager;
    def.listen_service = args[ARG_DPCAP_SERVER_PORT].s;
    def.send_thresh = args[ARG_DPCAP_SERVER_PACKET_BUFFER].i;
    KLTimingSetDouble(&def.send_interval,
		    args[ARG_DPCAP_SERVER_SEND_TIMEOUT].d);
    def.filter_cmd = args[ARG_DPCAP_FILTER].s;
    def.cap_len = args[ARG_DPCAP_CAP_LEN].i;
    def.verbosity = args[ARG_VERBOSE].i; 
    def.init_finished = n_server_init_finished;
    def.init_finished_client_data = NULL;

    if (args[ARG_READ_TRACE].s[0] != '\0') {
	/*
	 * Allow reading from traces at servers to test client-server
	 * interaction.
	 */
	def.type = DPCapServerTypeTrace;
	def.type_data.trace_name = args[ARG_READ_TRACE].s;
    }
    else {
	def.type = DPCapServerTypeLive;
	def.type_data.interface = args[ARG_INTERFACE].s;
    }
    ex = DPCapServerNew(&def, server_p);
    return ex;
}

static KLException *
n_event_handler_sigint(const KLEvent *event)
{
    assert(event != NULL);
    /* This is useful to see how memory is being used during a run. */
#if 0
    dmalloc_log_unfreed();
#endif
    return KLExceptionThrow(n_caught_sigint_ec);
}

static KLEventManager *
n_event_manager_init(KLParseOptionValue *args)
{
    KLEventManager *manager;
    KLEventDef def;
    KLEventDefRegistration *reg;
    KLEventManagerDef manager_def;

    /* Init event manager. */
    manager_def.read_log_fd = -1;
    manager_def.write_log_fd = -1;
    if (args[ARG_EVENT_LOG].s[0] != '\0') {
	assert((manager_def.write_log_fd = open(
	    args[ARG_EVENT_LOG].s, O_CREAT|O_TRUNC|O_WRONLY, 
	    S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH)) >= 0);
	assert(ftruncate(manager_def.write_log_fd, 0) >= 0);
    }
    KLExceptionErrorCatch(KLEventManagerNew(&manager_def, &manager), "");

    /* Register an event for SIGINT. */
    def.type = KLEventTypeSignal;
    def.type_data.signo = SIGINT;
    def.client_data_delete_fun = NULL;
    def.client_data = NULL;
    def.handler_fun = &n_event_handler_sigint;
    reg = KLEventDefRegister(manager, &def);

    return manager;
}

int
main(int argc, char * argv[])
{
    KLParseOptionValue *args;
    KLEventManager *manager = NULL;
    KLException *ex = NULL;
    const KLExceptionClass *ex_class;
    KLExceptionFlag flags;
    DPCapServer *server = NULL;
    Client *client = NULL;
    struct timespec timeout;
    int run_server, rval;

    /* Initialize and get command line arguments. */
    ex = n_parse_command_line(argc, argv, &args, &run_server);
    if (ex != NULL) {
	goto done;
    }
    if (args[ARG_VERSION].i) {
	goto done;
    }
    manager = n_event_manager_init(args);

    if (run_server) {
	ex = n_dpcap_server_init(manager, args, &server);
    }
    else {
	ex = client_new(manager, args, &client);
    }

    if (ex != NULL) {
	goto done;
    }

    KLTimingSet(&timeout, 60, 0);

    while (ex == NULL) {
	ex = KLEventDoNext(manager, LONG_MAX, &timeout);
    }

 done:
    if (server != NULL) {
	DPCapServerDelete(server);
    }
    if (client != NULL) {
	client_delete(client);
    }
    if (manager != NULL) {
	KLEventManagerDelete(manager);
    }
    if (args != NULL) {
	KLParseOptionFree(nettimer_option_specs, args);
    }
    rval = 0;
    if (ex != NULL) {
	flags = KLExceptionFlagPrint;
	ex_class = KLExceptionGetClass(ex);

	/* Catch exceptions for successful exit. */
	if ((ex_class == &nettimer_quit_ec) ||
	    (ex_class == &active_probing_below_max_error_ec) ||
	    (ex_class == &DPCapClientAllTracesFinishedEC)) {
	    flags = 0;
	}
	/* Catch help exception for printing to stdout. */
	else if (ex_class == &KLParseOptionECHelp) {
	    printf("%s\n", KLExceptionGetMeaning(ex));
	    flags = 0;
	}
	else {
	    rval = 1;
	}
	KLExceptionHandle(flags, ex, "nettimer exitting\n");
    }

    return rval;
}

